@@ -1,16 +1,17 @@
-use crate::data::{ComponentSetMut, ComponentSetOption};
 use crate::dynamics::MassProperties;
 use crate::geometry::{
     ColliderChanges, ColliderHandle, ColliderMassProps, ColliderParent, ColliderPosition,
-    ColliderShape,
+    ColliderSet, ColliderShape,
+};
+use crate::math::{
+    AngVector, AngularInertia, Isometry, Point, Real, Rotation, Translation, Vector,
 };
-use crate::math::{AngVector, AngularInertia, Isometry, Point, Real, Translation, Vector};
 use crate::parry::partitioning::IndexedData;
-use crate::utils::{WCross, WDot};
+use crate::utils::{WAngularInertia, WCross, WDot};
 use num::Zero;
 
 /// The unique handle of a rigid body added to a `RigidBodySet`.
-#[derive(Copy, Clone, Debug, PartialEq, Eq, Hash)]
+#[derive(Copy, Clone, Debug, PartialEq, Eq, Hash, Default)]
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
 #[repr(transparent)]
 pub struct RigidBodyHandle(pub crate::data::arena::Index);
@@ -55,8 +56,8 @@ pub type BodyStatus = RigidBodyType;
 pub enum RigidBodyType {
     /// A `RigidBodyType::Dynamic` body can be affected by all external forces.
     Dynamic = 0,
-    /// A `RigidBodyType::Static` body cannot be affected by external forces.
-    Static = 1,
+    /// A `RigidBodyType::Fixed` body cannot be affected by external forces.
+    Fixed = 1,
     /// A `RigidBodyType::KinematicPositionBased` body cannot be affected by any external forces but can be controlled
     /// by the user at the position level while keeping realistic one-way interaction with dynamic bodies.
     ///
@@ -71,14 +72,14 @@ pub enum RigidBodyType {
     /// cannot be pushed by anything. In other words, the trajectory of a kinematic body can only be
     /// modified by the user and is independent from any contact or joint it is involved in.
     KinematicVelocityBased = 3,
-    // Semikinematic, // A kinematic that performs automatic CCD with the static environment to avoid traversing it?
+    // Semikinematic, // A kinematic that performs automatic CCD with the fixed environment to avoid traversing it?
     // Disabled,
 }
 
 impl RigidBodyType {
-    /// Is this rigid-body static (i.e. cannot move)?
-    pub fn is_static(self) -> bool {
-        self == RigidBodyType::Static
+    /// Is this rigid-body fixed (i.e. cannot move)?
+    pub fn is_fixed(self) -> bool {
+        self == RigidBodyType::Fixed
     }
 
     /// Is this rigid-body dynamic (i.e. can move and be affected by forces)?
@@ -107,6 +108,8 @@ bitflags::bitflags! {
         const COLLIDERS   = 1 << 3;
         /// Flag indicating that the `RigidBodyType` component of this rigid-body has been modified.
         const TYPE        = 1 << 4;
+        /// Flag indicating that the `RigidBodyDominance` component of this rigid-body has been modified.
+        const DOMINANCE   = 1 << 5;
     }
 }
 
@@ -117,7 +120,7 @@ impl Default for RigidBodyChanges {
 }
 
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
-#[derive(Clone, Debug, Copy)]
+#[derive(Clone, Debug, Copy, PartialEq)]
 /// The position of this rigid-body.
 pub struct RigidBodyPosition {
     /// The world-space position of the rigid-body.
@@ -147,8 +150,11 @@ impl RigidBodyPosition {
     /// Computes the velocity need to travel from `self.position` to `self.next_position` in
     /// a time equal to `1.0 / inv_dt`.
     #[must_use]
-    pub fn interpolate_velocity(&self, inv_dt: Real) -> RigidBodyVelocity {
-        let dpos = self.next_position * self.position.inverse();
+    pub fn interpolate_velocity(&self, inv_dt: Real, local_com: &Point<Real>) -> RigidBodyVelocity {
+        let com = self.position * local_com;
+        let shift = Translation::from(com.coords);
+        let dpos = shift.inverse() * self.next_position * self.position.inverse() * shift;
+
         let angvel;
         #[cfg(feature = "dim2")]
         {
@@ -159,6 +165,7 @@ impl RigidBodyPosition {
             angvel = dpos.rotation.scaled_axis() * inv_dt;
         }
         let linvel = dpos.translation.vector * inv_dt;
+
         RigidBodyVelocity { linvel, angvel }
     }
 
@@ -194,15 +201,22 @@ where
 bitflags::bitflags! {
     #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
     /// Flags affecting the behavior of the constraints solver for a given contact manifold.
-    pub struct RigidBodyMassPropsFlags: u8 {
+    // FIXME: rename this to LockedAxes
+    pub struct LockedAxes: u8 {
+        /// Flag indicating that the rigid-body cannot translate along the `X` axis.
+        const TRANSLATION_LOCKED_X = 1 << 0;
+        /// Flag indicating that the rigid-body cannot translate along the `Y` axis.
+        const TRANSLATION_LOCKED_Y = 1 << 1;
+        /// Flag indicating that the rigid-body cannot translate along the `Z` axis.
+        const TRANSLATION_LOCKED_Z = 1 << 2;
         /// Flag indicating that the rigid-body cannot translate along any direction.
-        const TRANSLATION_LOCKED = 1 << 0;
+        const TRANSLATION_LOCKED = Self::TRANSLATION_LOCKED_X.bits | Self::TRANSLATION_LOCKED_Y.bits | Self::TRANSLATION_LOCKED_Z.bits;
         /// Flag indicating that the rigid-body cannot rotate along the `X` axis.
-        const ROTATION_LOCKED_X = 1 << 1;
-        /// Flag indicating that the rigid-body cannot rotate along the `X` axis.
-        const ROTATION_LOCKED_Y = 1 << 2;
-        /// Flag indicating that the rigid-body cannot rotate along the `X` axis.
-        const ROTATION_LOCKED_Z = 1 << 3;
+        const ROTATION_LOCKED_X = 1 << 3;
+        /// Flag indicating that the rigid-body cannot rotate along the `Y` axis.
+        const ROTATION_LOCKED_Y = 1 << 4;
+        /// Flag indicating that the rigid-body cannot rotate along the `Z` axis.
+        const ROTATION_LOCKED_Z = 1 << 5;
         /// Combination of flags indicating that the rigid-body cannot rotate along any axis.
         const ROTATION_LOCKED = Self::ROTATION_LOCKED_X.bits | Self::ROTATION_LOCKED_Y.bits | Self::ROTATION_LOCKED_Z.bits;
     }
@@ -210,17 +224,19 @@ bitflags::bitflags! {
 
 // TODO: split this into "LocalMassProps" and `WorldMassProps"?
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
-#[derive(Clone, Debug, Copy)]
+#[derive(Clone, Debug, PartialEq)]
 /// The mass properties of this rigid-bodies.
 pub struct RigidBodyMassProps {
     /// Flags for locking rotation and translation.
-    pub flags: RigidBodyMassPropsFlags,
+    pub flags: LockedAxes,
     /// The local mass properties of the rigid-body.
     pub local_mprops: MassProperties,
+    /// Mass-properties of this rigid-bodies, added to the contributions of its attached colliders.
+    pub additional_local_mprops: Option<Box<MassProperties>>,
     /// The world-space center of mass of the rigid-body.
     pub world_com: Point<Real>,
     /// The inverse mass taking into account translation locking.
-    pub effective_inv_mass: Real,
+    pub effective_inv_mass: Vector<Real>,
     /// The square-root of the world-space inverse angular inertia tensor of the rigid-body,
     /// taking into account rotation locking.
     pub effective_world_inv_inertia_sqrt: AngularInertia<Real>,
@@ -229,17 +245,18 @@ pub struct RigidBodyMassProps {
 impl Default for RigidBodyMassProps {
     fn default() -> Self {
         Self {
-            flags: RigidBodyMassPropsFlags::empty(),
+            flags: LockedAxes::empty(),
             local_mprops: MassProperties::zero(),
+            additional_local_mprops: None,
             world_com: Point::origin(),
-            effective_inv_mass: 0.0,
+            effective_inv_mass: Vector::zero(),
             effective_world_inv_inertia_sqrt: AngularInertia::zero(),
         }
     }
 }
 
-impl From<RigidBodyMassPropsFlags> for RigidBodyMassProps {
-    fn from(flags: RigidBodyMassPropsFlags) -> Self {
+impl From<LockedAxes> for RigidBodyMassProps {
+    fn from(flags: LockedAxes) -> Self {
         Self {
             flags,
             ..Self::default()
@@ -266,57 +283,86 @@ impl RigidBodyMassProps {
     /// The effective mass (that takes the potential translation locking into account) of
     /// this rigid-body.
     #[must_use]
-    pub fn effective_mass(&self) -> Real {
-        crate::utils::inv(self.effective_inv_mass)
+    pub fn effective_mass(&self) -> Vector<Real> {
+        self.effective_inv_mass.map(crate::utils::inv)
+    }
+
+    /// The effective world-space angular inertia (that takes the potential rotation locking into account) of
+    /// this rigid-body.
+    #[must_use]
+    pub fn effective_angular_inertia(&self) -> AngularInertia<Real> {
+        self.effective_world_inv_inertia_sqrt.squared().inverse()
+    }
+
+    /// Recompute the mass-properties of this rigid-bodies based on its currentely attached colliders.
+    pub fn recompute_mass_properties_from_colliders(
+        &mut self,
+        colliders: &ColliderSet,
+        attached_colliders: &RigidBodyColliders,
+        position: &Isometry<Real>,
+    ) {
+        self.local_mprops = self
+            .additional_local_mprops
+            .as_ref()
+            .map(|mprops| **mprops)
+            .unwrap_or_else(MassProperties::default);
+
+        for handle in &attached_colliders.0 {
+            if let Some(co) = colliders.get(*handle) {
+                if let Some(co_parent) = co.parent {
+                    let to_add = co
+                        .mprops
+                        .mass_properties(&*co.shape)
+                        .transform_by(&co_parent.pos_wrt_parent);
+                    self.local_mprops += to_add;
+                }
+            }
+        }
+
+        self.update_world_mass_properties(position);
     }
 
     /// Update the world-space mass properties of `self`, taking into account the new position.
     pub fn update_world_mass_properties(&mut self, position: &Isometry<Real>) {
         self.world_com = self.local_mprops.world_com(&position);
-        self.effective_inv_mass = self.local_mprops.inv_mass;
+        self.effective_inv_mass = Vector::repeat(self.local_mprops.inv_mass);
         self.effective_world_inv_inertia_sqrt =
             self.local_mprops.world_inv_inertia_sqrt(&position.rotation);
 
         // Take into account translation/rotation locking.
-        if self
-            .flags
-            .contains(RigidBodyMassPropsFlags::TRANSLATION_LOCKED)
-        {
-            self.effective_inv_mass = 0.0;
+        if self.flags.contains(LockedAxes::TRANSLATION_LOCKED_X) {
+            self.effective_inv_mass.x = 0.0;
+        }
+
+        if self.flags.contains(LockedAxes::TRANSLATION_LOCKED_Y) {
+            self.effective_inv_mass.y = 0.0;
+        }
+
+        #[cfg(feature = "dim3")]
+        if self.flags.contains(LockedAxes::TRANSLATION_LOCKED_Z) {
+            self.effective_inv_mass.z = 0.0;
         }
 
         #[cfg(feature = "dim2")]
         {
-            if self
-                .flags
-                .contains(RigidBodyMassPropsFlags::ROTATION_LOCKED_Z)
-            {
+            if self.flags.contains(LockedAxes::ROTATION_LOCKED_Z) {
                 self.effective_world_inv_inertia_sqrt = 0.0;
             }
         }
         #[cfg(feature = "dim3")]
         {
-            if self
-                .flags
-                .contains(RigidBodyMassPropsFlags::ROTATION_LOCKED_X)
-            {
+            if self.flags.contains(LockedAxes::ROTATION_LOCKED_X) {
                 self.effective_world_inv_inertia_sqrt.m11 = 0.0;
                 self.effective_world_inv_inertia_sqrt.m12 = 0.0;
                 self.effective_world_inv_inertia_sqrt.m13 = 0.0;
             }
 
-            if self
-                .flags
-                .contains(RigidBodyMassPropsFlags::ROTATION_LOCKED_Y)
-            {
+            if self.flags.contains(LockedAxes::ROTATION_LOCKED_Y) {
                 self.effective_world_inv_inertia_sqrt.m22 = 0.0;
                 self.effective_world_inv_inertia_sqrt.m12 = 0.0;
                 self.effective_world_inv_inertia_sqrt.m23 = 0.0;
             }
-            if self
-                .flags
-                .contains(RigidBodyMassPropsFlags::ROTATION_LOCKED_Z)
-            {
+            if self.flags.contains(LockedAxes::ROTATION_LOCKED_Z) {
                 self.effective_world_inv_inertia_sqrt.m33 = 0.0;
                 self.effective_world_inv_inertia_sqrt.m13 = 0.0;
                 self.effective_world_inv_inertia_sqrt.m23 = 0.0;
@@ -326,7 +372,7 @@ impl RigidBodyMassProps {
 }
 
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
-#[derive(Clone, Debug, Copy)]
+#[derive(Clone, Debug, Copy, PartialEq)]
 /// The velocities of this rigid-body.
 pub struct RigidBodyVelocity {
     /// The linear velocity of the rigid-body.
@@ -342,6 +388,38 @@ impl Default for RigidBodyVelocity {
 }
 
 impl RigidBodyVelocity {
+    /// Create a new rigid-body velocity component.
+    #[must_use]
+    pub fn new(linvel: Vector<Real>, angvel: AngVector<Real>) -> Self {
+        Self { linvel, angvel }
+    }
+
+    /// Converts a slice to a rigid-body velocity.
+    ///
+    /// The slice must contain at least 3 elements: the `slice[0..2]  contains
+    /// the linear velocity and the `slice[2]` contains the angular velocity.
+    #[must_use]
+    #[cfg(feature = "dim2")]
+    pub fn from_slice(slice: &[Real]) -> Self {
+        Self {
+            linvel: Vector::new(slice[0], slice[1]),
+            angvel: slice[2],
+        }
+    }
+
+    /// Converts a slice to a rigid-body velocity.
+    ///
+    /// The slice must contain at least 6 elements: the `slice[0..3]  contains
+    /// the linear velocity and the `slice[3..6]` contains the angular velocity.
+    #[must_use]
+    #[cfg(feature = "dim3")]
+    pub fn from_slice(slice: &[Real]) -> Self {
+        Self {
+            linvel: Vector::new(slice[0], slice[1], slice[2]),
+            angvel: AngVector::new(slice[3], slice[4], slice[5]),
+        }
+    }
+
     /// Velocities set to zero.
     #[must_use]
     pub fn zero() -> Self {
@@ -351,6 +429,70 @@ impl RigidBodyVelocity {
         }
     }
 
+    /// This velocity seen as a slice.
+    ///
+    /// The linear part is stored first.
+    #[inline]
+    pub fn as_slice(&self) -> &[Real] {
+        self.as_vector().as_slice()
+    }
+
+    /// This velocity seen as a mutable slice.
+    ///
+    /// The linear part is stored first.
+    #[inline]
+    pub fn as_mut_slice(&mut self) -> &mut [Real] {
+        self.as_vector_mut().as_mut_slice()
+    }
+
+    /// This velocity seen as a vector.
+    ///
+    /// The linear part is stored first.   
+    #[inline]
+    #[cfg(feature = "dim2")]
+    pub fn as_vector(&self) -> &na::Vector3<Real> {
+        unsafe { std::mem::transmute(self) }
+    }
+
+    /// This velocity seen as a mutable vector.
+    ///
+    /// The linear part is stored first.    
+    #[inline]
+    #[cfg(feature = "dim2")]
+    pub fn as_vector_mut(&mut self) -> &mut na::Vector3<Real> {
+        unsafe { std::mem::transmute(self) }
+    }
+
+    /// This velocity seen as a vector.
+    ///
+    /// The linear part is stored first.   
+    #[inline]
+    #[cfg(feature = "dim3")]
+    pub fn as_vector(&self) -> &na::Vector6<Real> {
+        unsafe { std::mem::transmute(self) }
+    }
+
+    /// This velocity seen as a mutable vector.
+    ///
+    /// The linear part is stored first.    
+    #[inline]
+    #[cfg(feature = "dim3")]
+    pub fn as_vector_mut(&mut self) -> &mut na::Vector6<Real> {
+        unsafe { std::mem::transmute(self) }
+    }
+
+    /// Return `self` rotated by `rotation`.
+    #[must_use]
+    pub fn transformed(self, rotation: &Rotation<Real>) -> Self {
+        Self {
+            linvel: rotation * self.linvel,
+            #[cfg(feature = "dim2")]
+            angvel: self.angvel,
+            #[cfg(feature = "dim3")]
+            angvel: rotation * self.angvel,
+        }
+    }
+
     /// The approximate kinetic energy of this rigid-body.
     ///
     /// This approximation does not take the rigid-body's mass and angular inertia
@@ -425,7 +567,7 @@ impl RigidBodyVelocity {
     /// The impulse is applied right away, changing the linear velocity.
     /// This does nothing on non-dynamic bodies.
     pub fn apply_impulse(&mut self, rb_mprops: &RigidBodyMassProps, impulse: Vector<Real>) {
-        self.linvel += impulse * rb_mprops.effective_inv_mass;
+        self.linvel += impulse.component_mul(&rb_mprops.effective_inv_mass);
     }
 
     /// Applies an angular impulse at the center-of-mass of this rigid-body.
@@ -465,8 +607,40 @@ impl RigidBodyVelocity {
     }
 }
 
+impl std::ops::Mul<Real> for RigidBodyVelocity {
+    type Output = Self;
+
+    #[must_use]
+    fn mul(self, rhs: Real) -> Self {
+        RigidBodyVelocity {
+            linvel: self.linvel * rhs,
+            angvel: self.angvel * rhs,
+        }
+    }
+}
+
+impl std::ops::Add<RigidBodyVelocity> for RigidBodyVelocity {
+    type Output = Self;
+
+    #[must_use]
+    fn add(self, rhs: Self) -> Self {
+        RigidBodyVelocity {
+            linvel: self.linvel + rhs.linvel,
+            angvel: self.angvel + rhs.angvel,
+        }
+    }
+}
+
+impl std::ops::AddAssign<RigidBodyVelocity> for RigidBodyVelocity {
+    #[must_use]
+    fn add_assign(&mut self, rhs: Self) {
+        self.linvel += rhs.linvel;
+        self.angvel += rhs.angvel;
+    }
+}
+
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
-#[derive(Clone, Debug, Copy)]
+#[derive(Clone, Debug, Copy, PartialEq)]
 /// Damping factors to progressively slow down a rigid-body.
 pub struct RigidBodyDamping {
     /// Damping factor for gradually slowing down the translational motion of the rigid-body.
@@ -485,7 +659,7 @@ impl Default for RigidBodyDamping {
 }
 
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
-#[derive(Clone, Debug, Copy)]
+#[derive(Clone, Debug, Copy, PartialEq)]
 /// The user-defined external forces applied to this rigid-body.
 pub struct RigidBodyForces {
     /// Accumulation of external forces (only for dynamic bodies).
@@ -495,6 +669,10 @@ pub struct RigidBodyForces {
     /// Gravity is multiplied by this scaling factor before it's
     /// applied to this rigid-body.
     pub gravity_scale: Real,
+    /// Forces applied by the user.
+    pub user_force: Vector<Real>,
+    /// Torque applied by the user.
+    pub user_torque: AngVector<Real>,
 }
 
 impl Default for RigidBodyForces {
@@ -503,6 +681,8 @@ impl Default for RigidBodyForces {
             force: na::zero(),
             torque: na::zero(),
             gravity_scale: 1.0,
+            user_force: na::zero(),
+            user_torque: na::zero(),
         }
     }
 }
@@ -516,7 +696,7 @@ impl RigidBodyForces {
         init_vels: &RigidBodyVelocity,
         mprops: &RigidBodyMassProps,
     ) -> RigidBodyVelocity {
-        let linear_acc = self.force * mprops.effective_inv_mass;
+        let linear_acc = self.force.component_mul(&mprops.effective_inv_mass);
         let angular_acc = mprops.effective_world_inv_inertia_sqrt
             * (mprops.effective_world_inv_inertia_sqrt * self.torque);
 
@@ -528,8 +708,13 @@ impl RigidBodyForces {
 
     /// Adds to `self` the gravitational force that would result in a gravitational acceleration
     /// equal to `gravity`.
-    pub fn add_gravity_acceleration(&mut self, gravity: &Vector<Real>, mass: Real) {
-        self.force += gravity * self.gravity_scale * mass;
+    pub fn compute_effective_force_and_torque(
+        &mut self,
+        gravity: &Vector<Real>,
+        mass: &Vector<Real>,
+    ) {
+        self.force = self.user_force + gravity.component_mul(&mass) * self.gravity_scale;
+        self.torque = self.user_torque;
     }
 
     /// Applies a force at the given world-space point of the rigid-body with the given mass properties.
@@ -539,13 +724,13 @@ impl RigidBodyForces {
         force: Vector<Real>,
         point: Point<Real>,
     ) {
-        self.force += force;
-        self.torque += (point - rb_mprops.world_com).gcross(force);
+        self.user_force += force;
+        self.user_torque += (point - rb_mprops.world_com).gcross(force);
     }
 }
 
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
-#[derive(Clone, Debug, Copy)]
+#[derive(Clone, Debug, Copy, PartialEq)]
 /// Information used for Continuous-Collision-Detection.
 pub struct RigidBodyCcd {
     /// The distance used by the CCD solver to decide if a movement would
@@ -617,7 +802,7 @@ impl RigidBodyCcd {
 }
 
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
-#[derive(Clone, Debug, Copy)]
+#[derive(Clone, Debug, Copy, PartialEq, Eq, Hash)]
 /// Internal identifiers used by the physics engine.
 pub struct RigidBodyIds {
     pub(crate) active_island_id: usize,
@@ -638,7 +823,7 @@ impl Default for RigidBodyIds {
 }
 
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
-#[derive(Clone, Debug)]
+#[derive(Clone, Debug, PartialEq, Eq)]
 /// The set of colliders attached to this rigid-bodies.
 ///
 /// This should not be modified manually unless you really know what
@@ -703,39 +888,31 @@ impl RigidBodyColliders {
     }
 
     /// Update the positions of all the colliders attached to this rigid-body.
-    pub fn update_positions<Colliders>(
+    pub fn update_positions(
         &self,
-        colliders: &mut Colliders,
+        colliders: &mut ColliderSet,
         modified_colliders: &mut Vec<ColliderHandle>,
         parent_pos: &Isometry<Real>,
-    ) where
-        Colliders: ComponentSetMut<ColliderPosition>
-            + ComponentSetMut<ColliderChanges>
-            + ComponentSetOption<ColliderParent>,
-    {
+    ) {
         for handle in &self.0 {
             // NOTE: the ColliderParent component must exist if we enter this method.
-            let co_parent: &ColliderParent = colliders
-                .get(handle.0)
-                .expect("Could not find the ColliderParent component.");
-            let new_pos = parent_pos * co_parent.pos_wrt_parent;
+            let co = colliders.index_mut_internal(*handle);
+            let new_pos = parent_pos * co.parent.as_ref().unwrap().pos_wrt_parent;
+
+            if !co.changes.contains(ColliderChanges::MODIFIED) {
+                modified_colliders.push(*handle);
+            }
 
             // Set the modification flag so we can benefit from the modification-tracking
             // when updating the narrow-phase/broad-phase afterwards.
-            colliders.map_mut_internal(handle.0, |co_changes: &mut ColliderChanges| {
-                if !co_changes.contains(ColliderChanges::MODIFIED) {
-                    modified_colliders.push(*handle);
-                }
-
-                *co_changes |= ColliderChanges::POSITION;
-            });
-            colliders.set_internal(handle.0, ColliderPosition(new_pos));
+            co.changes |= ColliderChanges::POSITION;
+            co.pos = ColliderPosition(new_pos);
         }
     }
 }
 
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
-#[derive(Clone, Debug, Copy)]
+#[derive(Clone, Debug, Copy, PartialEq, Eq, PartialOrd, Ord, Hash)]
 /// The dominance groups of a rigid-body.
 pub struct RigidBodyDominance(pub i8);
 
@@ -746,7 +923,7 @@ impl Default for RigidBodyDominance {
 }
 
 impl RigidBodyDominance {
-    /// The actually dominance group of this rigid-body, after taking into account its type.
+    /// The actual dominance group of this rigid-body, after taking into account its type.
     pub fn effective_group(&self, status: &RigidBodyType) -> i16 {
         if status.is_dynamic() {
             self.0 as i16
@@ -760,14 +937,16 @@ impl RigidBodyDominance {
 ///
 /// This controls whether a body is sleeping or not.
 /// If the threshold is negative, the body never sleeps.
-#[derive(Copy, Clone, Debug)]
+#[derive(Copy, Clone, Debug, PartialEq)]
 #[cfg_attr(feature = "serde-serialize", derive(Serialize, Deserialize))]
 pub struct RigidBodyActivation {
-    /// The threshold pseudo-kinetic energy bellow which the body can fall asleep.
-    pub threshold: Real,
-    /// The current pseudo-kinetic energy of the body.
-    pub energy: Real,
-    /// Is this body already sleeping?
+    /// The threshold linear velocity bellow which the body can fall asleep.
+    pub linear_threshold: Real,
+    /// The angular linear velocity bellow which the body can fall asleep.
+    pub angular_threshold: Real,
+    /// Since how much time can this body sleep?
+    pub time_since_can_sleep: Real,
+    /// Is this body sleeping?
     pub sleeping: bool,
 }
 
@@ -778,16 +957,28 @@ impl Default for RigidBodyActivation {
 }
 
 impl RigidBodyActivation {
-    /// The default amount of energy bellow which a body can be put to sleep by nphysics.
-    pub fn default_threshold() -> Real {
-        0.01
+    /// The default linear velocity bellow which a body can be put to sleep.
+    pub fn default_linear_threshold() -> Real {
+        0.4
+    }
+
+    /// The default angular velocity bellow which a body can be put to sleep.
+    pub fn default_angular_threshold() -> Real {
+        0.5
+    }
+
+    /// The amount of time the rigid-body must remain bellow it’s linear and angular velocity
+    /// threshold before falling to sleep.
+    pub fn default_time_until_sleep() -> Real {
+        2.0
     }
 
     /// Create a new rb_activation status initialised with the default rb_activation threshold and is active.
     pub fn active() -> Self {
         RigidBodyActivation {
-            threshold: Self::default_threshold(),
-            energy: Self::default_threshold() * 4.0,
+            linear_threshold: Self::default_linear_threshold(),
+            angular_threshold: Self::default_angular_threshold(),
+            time_since_can_sleep: 0.0,
             sleeping: false,
         }
     }
@@ -795,16 +986,18 @@ impl RigidBodyActivation {
     /// Create a new rb_activation status initialised with the default rb_activation threshold and is inactive.
     pub fn inactive() -> Self {
         RigidBodyActivation {
-            threshold: Self::default_threshold(),
-            energy: 0.0,
+            linear_threshold: Self::default_linear_threshold(),
+            angular_threshold: Self::default_angular_threshold(),
             sleeping: true,
+            time_since_can_sleep: Self::default_time_until_sleep(),
         }
     }
 
     /// Create a new activation status that prevents the rigid-body from sleeping.
     pub fn cannot_sleep() -> Self {
         RigidBodyActivation {
-            threshold: -Real::MAX,
+            linear_threshold: -1.0,
+            angular_threshold: -1.0,
             ..Self::active()
         }
     }
@@ -812,22 +1005,22 @@ impl RigidBodyActivation {
     /// Returns `true` if the body is not asleep.
     #[inline]
     pub fn is_active(&self) -> bool {
-        self.energy != 0.0
+        !self.sleeping
     }
 
     /// Wakes up this rigid-body.
     #[inline]
     pub fn wake_up(&mut self, strong: bool) {
         self.sleeping = false;
-        if strong || self.energy == 0.0 {
-            self.energy = self.threshold.abs() * 2.0;
+        if strong {
+            self.time_since_can_sleep = 0.0;
         }
     }
 
     /// Put this rigid-body to sleep.
     #[inline]
     pub fn sleep(&mut self) {
-        self.energy = 0.0;
         self.sleeping = true;
+        self.time_since_can_sleep = Self::default_time_until_sleep();
     }
 }
